/* $Id: perfctr.h,v 1.9 2001/06/10 22:36:40 mikpe Exp $
 * x86 Performance-Monitoring Counters driver
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */
#ifndef _ASM_I386_PERFCTR_H
#define _ASM_I386_PERFCTR_H

struct perfctr_sum_ctrs {
	unsigned long long tsc;
	unsigned long long pmc[18];
};

struct perfctr_low_ctrs {
	unsigned int tsc;
	unsigned int pmc[18];
};

struct perfctr_cpu_control_user {
	unsigned int tsc_on;
	unsigned int nractrs;		/* # of a-mode counters */
	unsigned int nrictrs;		/* # of i-mode counters */
	unsigned int pmc_map[18];
	unsigned int evntsel[18];	/* one per counter, even on P5 */
	unsigned int evntsel_aux[18];	/* e.g. P4 ESCR contents */
	int ireset[18];			/* <= 0, for i-mode counters */
};

struct perfctr_cpu_control {
	/* These two fields are visible to the user-space C compiler, as
	   part of the layout of mmap:ed vperfctrs, but are not part of
	   the user-kernel API. They are put here to ensure small structure
	   offsets for the most frequently used fields. */
	unsigned int cstatus;	/* XXX: could be useful to user-space */
	union {
		unsigned int p5_cesr;
		unsigned int id;	/* P6/K7 cache validity checking */
	} k1;
	/* The 'user' field is part of the user-kernel API. */
	struct perfctr_cpu_control_user user;
	/* Fields below are strictly private to the low-level driver. */
	union {
		unsigned int p4_escr_map[18];
	} k2;
};

#ifdef __KERNEL__

#if defined(CONFIG_PERFCTR) || defined(CONFIG_PERFCTR_MODULE)

/* called from arch/i386/kernel/setup.c:dodgy_tsc() */
extern void perfctr_dodgy_tsc(void);

/* Driver init/exit. */
extern int perfctr_cpu_init(void);
extern void perfctr_cpu_exit(void);

/* CPU type name. */
extern char *perfctr_cpu_name[];

/* Hardware reservation. */
extern const char *perfctr_cpu_reserve(const char *service);
extern void perfctr_cpu_release(const char *service);

/* Check that the user's control data is valid.
   Update the kernel's private control data.
   Returns < 0 if the control data is invalid. */
extern int perfctr_cpu_check_control(struct perfctr_cpu_control *control);

/* Update the control and event selection data. PRE: control->cstatus != 0. */
extern void perfctr_cpu_write_control(const struct perfctr_cpu_control *control);

/* Read and save current counter values. PRE: control->cstatus != 0. */
extern void perfctr_cpu_read_counters(const struct perfctr_cpu_control *control,
				      struct perfctr_low_ctrs *ctrs);

/* Process resume/suspend hooks for virtual perfctrs. */
static inline void perfctr_cpu_enable_rdpmc(void) { }
static inline void perfctr_cpu_disable_rdpmc(void) { }

typedef void (*perfctr_ihandler_t)(unsigned long pc);

#ifdef CONFIG_X86_LOCAL_APIC
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#include <asm/fixmap.h>
#include <asm/apic.h>
struct hw_interrupt_type;
#include <asm/hw_irq.h>
#ifdef LOCAL_PERFCTR_VECTOR
#define PERFCTR_INTERRUPT_SUPPORT 1
#endif
#endif
#endif

#if PERFCTR_INTERRUPT_SUPPORT
extern unsigned int apic_lvtpc_irqs[];
extern void perfctr_interrupt(void);
extern void perfctr_cpu_set_ihandler(perfctr_ihandler_t);
extern void perfctr_cpu_isuspend(const struct perfctr_cpu_control*);
extern void perfctr_cpu_iresume(const struct perfctr_cpu_control*,
				const struct perfctr_low_ctrs*);
#else
static inline void perfctr_cpu_set_ihandler(perfctr_ihandler_t x) { }
static inline void perfctr_cpu_isuspend(const struct perfctr_cpu_control *x) { }
static inline void perfctr_cpu_iresume(const struct perfctr_cpu_control *x,
				       const struct perfctr_low_ctrs *y) { }
#endif

/* cstatus is a re-encoding of control->tsc_on/nractrs/nrictrs
   which should provide less overhead for the common cases */

static inline
unsigned int perfctr_mk_cstatus(unsigned int tsc_on, unsigned int nractrs,
				unsigned int nrictrs)
{
	return (tsc_on<<31) | (nrictrs<<16) | (nractrs<<8) | (nractrs+nrictrs);
}

static inline unsigned int perfctr_cstatus_enabled(unsigned int cstatus)
{
	return cstatus;
}

static inline int perfctr_cstatus_has_tsc(unsigned int cstatus)
{
	return (int)cstatus < 0;	/* test and jump on sign */
}

static inline unsigned int perfctr_cstatus_nrctrs(unsigned int cstatus)
{
	return cstatus & 0x7F;		/* and with imm8 */
}

static inline unsigned int perfctr_cstatus_nractrs(unsigned int cstatus)
{
	return (cstatus >> 8) & 0x7F;
}

static inline unsigned int perfctr_cstatus_has_ictrs(unsigned int cstatus)
{
#if PERFCTR_INTERRUPT_SUPPORT
	return cstatus & (0x7F << 16);
#else
	return 0;
#endif
}

/*
 * %cr4 access functions
 * shared between x86.c, x86_setup.c, and x86_tests.c
 * (were once used implicitly by virtual.c too but not any more)
 */
static inline void write_cr4(unsigned int x)
{
	__asm__ __volatile__("movl %0,%%cr4" : : "r"(x));
}

static inline unsigned int read_cr4(void)
{
	unsigned int x;
	__asm__ __volatile__("movl %%cr4,%0" : "=r"(x));
	return x;
}

static inline void set_in_cr4_local(unsigned int mask)
{
	write_cr4(read_cr4() | mask);
}

static inline void clear_in_cr4_local(unsigned int mask)
{
	write_cr4(read_cr4() & ~mask);
}

#else	/* !CONFIG_PERFCTR */

static inline void perfctr_dodgy_tsc(void) { }

#endif	/* CONFIG_PERFCTR */

#endif	/* __KERNEL__ */

#endif	/* _ASM_I386_PERFCTR_H */
